package minicp.enigma;

import minicp.cp.Factory;
import minicp.engine.constraints.*;
import minicp.engine.core.BoolVar;
import minicp.engine.core.IntVar;
import minicp.engine.core.Solver;
import minicp.search.DFSearch;
import minicp.search.SearchStatistics;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

import static minicp.cp.BranchingScheme.and;
import static minicp.cp.BranchingScheme.firstFail;
import static minicp.cp.Factory.makeDfs;

public class Bombe {
    public static void main(String[] args) {

        Alphabet alphabet = new Alphabet("ACDEGHILMNORSTUW");
        int[][] perm = new int[][]{{5, 15, 7, 6, 3, 4, 13, 9, 12, 0, 11, 8, 2, 1, 10, 14},
                {13, 3, 9, 11, 14, 1, 6, 2, 8, 0, 4, 7, 12, 10, 5, 15},
                {6, 11, 12, 0, 13, 15, 8, 7, 2, 5, 1, 14, 3, 9, 10, 4}
        };

        int[] reflector = new int[]{9, 15, 13, 14, 10, 6, 5, 11, 12, 0, 4, 7, 8, 2, 3, 1};
        int[] plug_board = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17}; // This is not the real plug board
        int nplugs = 2;


        System.out.println("\nEXEMPLE 1\n");
        String cipher = "SRICOGSISNUGNHHDANALOTWTOOWCARRRENIHMUCNLLALODRDNIHILWILSMODTRTCOTSWTWADNCUEUIWCM";

        EnigmaEmulator machine = new EnigmaEmulator(alphabet, perm, plug_board, reflector);
        tryKeys(machine, cipher, "CONGRATULATION", nplugs);

        // COMPLETE MACHINE //

        System.out.println("\nEXEMPLE 2\n");
        alphabet = new Alphabet("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
        nplugs = 10;


        perm = new int[][]{
                {21, 18, 16, 7, 15, 10, 19, 23, 1, 11, 25, 2, 4, 14, 17, 3, 9, 0, 12, 24, 5, 22, 8, 13, 20, 6},
                {19, 6, 0, 14, 12, 2, 1, 8, 11, 21, 5, 16, 7, 10, 18, 17, 15, 25, 9, 4, 24, 22, 3, 13, 20, 23},
                {19, 15, 12, 5, 16, 21, 3, 8, 24, 20, 14, 23, 7, 0, 9, 17, 13, 22, 10, 6, 1, 25, 2, 11, 18, 4}
        };
        reflector = new int[]{12, 22, 5, 9, 8, 2, 14, 18, 4, 3, 17, 13, 0, 11, 6, 21, 23, 10, 7, 25, 24, 15, 1, 16, 20, 19};
        plug_board = new int[]{20, 1, 4, 3, 2, 11, 9, 16, 22, 6, 15, 5, 17, 13, 14, 10, 7, 12, 23, 19, 0, 24, 8, 18, 21, 25};
        int[] keys = new int[]{8, 4, 16};

        machine = new EnigmaEmulator(alphabet,
                perm,
                plug_board,
                reflector
        );
        machine.set_key(keys);

        String message = "THISMESSAGEWASENCODEDWITHTHEREALMILITARYENIGMAMACHINEANDSHOULDBEINDECIPHERABLEBUTWEMANAGETOBREAKIT";

        String cipher2 = machine.encrypt(message);
        tryKeys(machine, cipher2, "THISMESSAGEWAS", nplugs);

    }
    /*
    * This algorithm try all keys, and look for contradiction, given the crib and the symetry of the plugboard
    * For each valid key, it run the CP program with the key as a constant instead of a variable
    */
    public static void tryKeys(EnigmaEmulator machine, String ciphertext, String crib, int nplugs) {

        Alphabet alphabet = machine.get_alphabet();
        int M = machine.num_rotors();
        int N = alphabet.size();
        int K = crib.length();

        int[] id = new int[N];
        for (int i = 0; i < N; ++i) id[i] = i;

        char[] crib_char = crib.toCharArray();
        char[] code_char = ciphertext.substring(0, K).toCharArray();

        assert code_char.length == K;

        boolean[] in_crib = new boolean[N];
        Arrays.fill(in_crib, false);
        for (char c : crib_char) in_crib[alphabet.get_digit(c)] = true;
        for (char c : code_char) in_crib[alphabet.get_digit(c)] = true;

        int[][] neighbor_list = new int[N][];
        int[][] menu = new int[N][N];
        for (int i = 0; i < N; ++i) {
            Arrays.fill(menu[i], -1);
        }

        for (int i = 0; i < K; ++i) {
            int x = alphabet.get_digit(crib_char[i]);
            int y = alphabet.get_digit(code_char[i]);
            menu[x][y] = i;
            menu[y][x] = i;
        }

        for (int x = 0; x < N; ++x) {
            List<Integer> neighbors = new ArrayList<>();
            for (int y = 0; y < N; ++y) if (menu[x][y] >= 0) neighbors.add(y);

            neighbor_list[x] = new int[neighbors.size()];
            for (int i = 0; i < neighbors.size(); ++i) neighbor_list[x][i] = neighbors.get(i);
        }

        boolean[][] forbidden = new boolean[N][N];
        int[] possible = new int[N];
        int[] stecker = new int[N];

        int[] stack = new int[2 * N * N];

        int max_iterations = (int) Math.pow(N, M);
        int[] key = new int[M];
        int[] ikey = new int[M];

        int nkeys = 0;
        for (int iteration = 0; iteration < max_iterations; ++iteration) {
            if (iteration % 100 == 0) {
                System.out.printf("\b\b\b\b\b\b\b\b\b%-5d %2d%%", nkeys, (iteration * 100) / max_iterations);
                System.out.flush();
            }

            for (int i = 0; i < M; ++i) {
                key[i] = (iteration / (int) Math.pow(N, i)) % N;
            }

            boolean support = true;
            int diagonal = N;
            for (int i = 0; i < N; ++i) {
                Arrays.fill(forbidden[i], false);
                possible[i] = N;
            }

            for (int s = 0; s < N && support; ++s) {
                if (in_crib[s]) {
                    for (int t = 0; t < N; ++t) {
                        int top = 0;
                        stack[top++] = s;
                        stack[top++] = t;
                        stack[top++] = t;
                        stack[top++] = s;

                        int pos = 0;
                        Arrays.fill(stecker, -1);

                        while (pos < top && !forbidden[t][s]) {
                            int x = stack[pos++];
                            int y = stack[pos++];

                            if (!forbidden[x][y]) {
                                stecker[x] = y;
                                stecker[y] = x;
                                for (int j = 0; j < neighbor_list[x].length && support; ++j) {
                                    int cx = neighbor_list[x][j];


                                    for (int k = 0; k < M; ++k) {
                                        ikey[k] = (key[k] + menu[x][cx] / (int) Math.pow(N, k)) % N;
                                    }
                                    machine.set_key(ikey);

                                    int ey = machine.encryption(y);

                                    if (stecker[ey] < 0 && stecker[cx] < 0) {
                                        stecker[ey] = cx;
                                        stecker[cx] = ey;
                                        stack[top++] = cx;
                                        stack[top++] = ey;
                                        if (ey != cx) {
                                            stack[top++] = ey;
                                            stack[top++] = cx;
                                        }
                                    } else if (stecker[cx] != ey) {

                                        stack[top++] = cx;
                                        stack[top++] = ey;
                                        if (ey != cx) {
                                            stack[top++] = ey;
                                            stack[top++] = cx;
                                        }
                                        while (top > 0 && support) {
                                            int c1 = stack[--top];
                                            int c2 = stack[--top];
                                            if (!forbidden[c1][c2]) {
                                                forbidden[c1][c2] = true;
                                                if (--possible[c1] == 0) support = false;
                                            }
                                            if (c1 == c2)
                                                if (--diagonal < alphabet.size() - 2*nplugs) support = false;

                                        }
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            if (support) {
                ++nkeys;
                System.out.print("\b\b\b\b\b\b\b\b\bSTOP");
                for (int k : key) {
                    System.out.print(" " + k);
                }
                System.out.printf(" @%d\n", iteration);
                BombeEnigmaBreak(machine, ciphertext, crib, key, nplugs);
            }
        }
        System.out.println();
    }
    static public void BombeEnigmaBreak(EnigmaEmulator machine, String ciphertext, String crib, int[] key, int nplugs) {
        Alphabet alphabet = machine.get_alphabet();
        char[] code_char = ciphertext.toCharArray();
        char[] crib_char = crib.toCharArray();
        Solver cp = Factory.makeSolver(false);

        int[][] frotors = machine.get_forward_rotors();
        int[] reflector = machine.get_reflector();

        int N = alphabet.size();
        int M = frotors.length;
        int K = code_char.length;

        /*************** TODO *******************/
        int[][] offset = new int[M][];
        for(int j = 0; j < M; ++j) {
            offset[j] = new int[N];
            for (int i = 0; i < N; ++i) {
                offset[j][i] = (frotors[j][i] - i);
                if (offset[j][i] < 0) {
                    offset[j][i] = offset[j][i] + N;
                }
            }
        }

        int[][][] transpo_matrix = new int[M][][];
        for(int m = 0; m < M; ++m){
            transpo_matrix[m] = new int[N][];
            for(int i = 0; i < N; ++i){
                transpo_matrix[m][i] = new int[N];
                for(int j = 0; j < N; ++j){
                    transpo_matrix[m][i][j] = (j + offset[m][(j + i)%N])%N;
                }
            }
        }


        IntVar[][] text = new IntVar[2 * M + 4][];

        IntVar[] plug_board = Factory.makeIntVarArray(cp, N, N);
        IntVar[] loop = new IntVar[N];
        text[0] = Factory.makeIntVarArray(cp, K, N); // plain
        text[1] = Factory.makeIntVarArray(cp, K, N); // switched
        for (int j = 1; j <= M; ++j) {
            text[j+1] = new IntVar[K]; // forward
        }
        text[M + 2] = new IntVar[K]; // reflector
        for (int j = 1; j <= M; ++j) {
            text[M + 2 + j] = Factory.makeIntVarArray(cp, K, N); // backward
        }
        text[2 * M + 3] = Factory.makeIntVarArray(cp, K, N); // cipher


        //plugboard
        Factory.allDifferent(plug_board);
        for (int i = 0; i < N; ++i) {
            loop[i] = Factory.isEqual(plug_board[i], i);
        }
        Factory.sum(loop, N - 2 * nplugs);

        //plugboard symetry
        for (int i = 0; i < alphabet.size(); ++i) {
            for (int j = 0; j < alphabet.size(); ++j) {
                BoolVar equal = Factory.isEqual(plug_board[i], j);
                cp.post(new IsEqual(equal, plug_board[j], i));
            }
        }

        // Assignment code text
        for(int i = 0; i < K; ++i){
            text[2 * M + 3][i].assign(alphabet.get_digit(code_char[i]));
        }
        //assignement crib
        for (int i = 0; i < crib_char.length; ++i) {
            text[0][i].assign(alphabet.get_digit(crib_char[i]));
        }
        //plug board constraints
        for (int i = 0; i < K; ++i) {
            cp.post(new Element1DVar(plug_board, text[0][i], text[1][i]));
            cp.post(new Element1DVar(plug_board, text[2 * M + 2][i], text[2 * M + 3][i]));
        }

        for (int i = 0; i < K; ++i) {
            for(int j = 0; j < M; ++j){
                int value = (int) (i / Math.pow(N, j)) + key[j];
                IntVar idx = Factory.modulo(Factory.makeIntVar(cp, value, value), N);
                text[j+1+1][i] = Factory.element(transpo_matrix[j], idx, text[j+1][i]);
            }
            //reflector
            text[M + 1+1][i] = Factory.element(reflector, text[M+1][i]);
            for(int j = 0; j < M; ++j){
                int rotor_idx = M - (j + 1);
                int value = (int) (i / Math.pow(N, rotor_idx)) + key[rotor_idx];
                IntVar idx = Factory.modulo(Factory.makeIntVar(cp, value, value), N);
                cp.post(new Element2D(transpo_matrix[rotor_idx], idx, text[M + 2 + j+1][i], text[M + 1 + j+1][i]));
            }
        }

        // implied constraints, only a few nodes are avoided with them, not a good tradeof
        ///*
        for (int i = 0; i < code_char.length; ++i) {
            for (int j = i + 1; j < code_char.length; ++j) {
                if(i < crib_char.length && j < crib_char.length){
                    if (crib_char[i] == crib_char[j])
                        cp.post(new Equal(text[1][i], text[1][j] ));
                    else
                        cp.post(new NotEqual(text[1][i], text[1][j]));
                }
                if(i < crib_char.length) {
                    if (crib_char[i] == code_char[j])
                        cp.post(new Equal(text[1][i], text[2 * M + 2][j]));
                    else
                        cp.post(new NotEqual(text[1][i], text[2 * M + 2][j]));
                }
                if(j < crib_char.length) {
                    if (code_char[i] == crib_char[j])
                        cp.post(new Equal(text[2 * M + 2][i], text[1][j]));
                    else
                        cp.post(new NotEqual(text[2 * M + 2][i], text[1][j]));
                }

                if (code_char[i] == code_char[j])
                    cp.post(new Equal(text[2 * M + 2][j], text[2 * M + 2][i]));
                else
                    cp.post(new NotEqual(text[2 * M + 2][j], text[2 * M + 2][i]));
            }
        }
        //*/

        DFSearch dfs = makeDfs(cp, and(firstFail(plug_board), firstFail(text[0])));

        IntVar[] plaintext = text[0];
        /*****************************************/

        dfs.onSolution(() -> {
                    System.out.print(" -");
                    for (int i = 0; i < N; ++i) {
                        System.out.print(plug_board[i]);
                    }
                    System.out.print(" / ");
                    for (int i = 0; i < M; ++i) {
                        System.out.print("{"+key[i]+"}");
                    }
                    System.out.print(" -> ");
                    alphabet.print(plaintext);
                    System.out.println();
                }
        );

        SearchStatistics stats = dfs.solve();
        System.out.println("CP search done");
    }
}
