package minicp.enigma;

import minicp.cp.Factory;
import minicp.engine.constraints.*;
import minicp.engine.core.BoolVar;
import minicp.engine.core.IntVar;
import minicp.engine.core.Solver;
import minicp.search.DFSearch;
import minicp.search.SearchStatistics;

import static minicp.cp.BranchingScheme.and;
import static minicp.cp.BranchingScheme.firstFail;
import static minicp.cp.Factory.makeDfs;

public class MilitaryEnigmaBreaker {

    public static void main(String[] args) {
        Alphabet alphabet = new Alphabet("ACDEGHILMNORSTUW");
        int[][] perm = new int[][]{{5, 15, 7, 6, 3, 4, 13, 9, 12, 0, 11, 8, 2, 1, 10, 14},
                {13, 3, 9, 11, 14, 1, 6, 2, 8, 0, 4, 7, 12, 10, 5, 15},
                {6, 11, 12, 0, 13, 15, 8, 7, 2, 5, 1, 14, 3, 9, 10, 4}
        };

        int[] reflector = new int[]{9, 15, 13, 14, 10, 6, 5, 11, 12, 0, 4, 7, 8, 2, 3, 1};
        int[] plug_board = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17}; // This is not the real plug board
        int nplugs = 2;


        System.out.println("\nEXERCICE 1\n");
        String cipher2 = "SRICOGSISNUGNHHDANALOTWTOOWCARRRENIHMUCNLLALODRDNIHILWILSMODTRTCOTSWTWADNCUEUIWCM";

        EnigmaEmulator machine = new EnigmaEmulator(alphabet, perm, plug_board, reflector);

        militaryEnigmaBreak(machine, nplugs, cipher2, "CONGRATULATION");

    }

    static public void militaryEnigmaBreak(EnigmaEmulator machine, int nplugs, String ciphertext, String crib) {
        Alphabet alphabet = machine.get_alphabet();
        char[] code_char = ciphertext.toCharArray();
        char[] crib_char = crib.toCharArray();
        Solver cp = Factory.makeSolver(false);

        int[][] frotors = machine.get_forward_rotors();
        int[] reflector = machine.get_reflector();

        int N = alphabet.size();
        int M = frotors.length;
        int K = code_char.length;

        /*************** TODO *******************/
        int[][] offset = new int[M][];
        for(int j = 0; j < M; ++j) {
            offset[j] = new int[N];
            for (int i = 0; i < N; ++i) {
                offset[j][i] = (frotors[j][i] - i);
                if (offset[j][i] < 0) {
                    offset[j][i] = offset[j][i] + N;
                }
            }
        }

        int[][][] transpo_matrix = new int[M][][];
        for(int m = 0; m < M; ++m){
            transpo_matrix[m] = new int[N][];
            for(int i = 0; i < N; ++i){
                transpo_matrix[m][i] = new int[N];
                for(int j = 0; j < N; ++j){
                    transpo_matrix[m][i][j] = (j + offset[m][(j + i)%N])%N;
                }
            }
        }


        IntVar[][] text = new IntVar[2 * M + 4][];

        IntVar[] plug_board = Factory.makeIntVarArray(cp, N, N);
        IntVar[] loop = new IntVar[N];
        IntVar[] key = Factory.makeIntVarArray(cp, M, N);

        text[0] = Factory.makeIntVarArray(cp, K, N); // plain
        text[1] = Factory.makeIntVarArray(cp, K, N); // switched
        for (int j = 1; j <= M; ++j) {
            text[j+1] = new IntVar[K]; // forward
        }
        text[M + 2] = new IntVar[K]; // reflector
        for (int j = 1; j <= M; ++j) {
            text[M + 2 + j] = Factory.makeIntVarArray(cp, K, N); // backward
        }
        text[2 * M + 3] = Factory.makeIntVarArray(cp, K, N); // cipher


        //plugboard
        Factory.allDifferent(plug_board);
        for (int i = 0; i < N; ++i) {
            loop[i] = Factory.isEqual(plug_board[i], i);
        }
        Factory.sum(loop, N - 2 * nplugs);

        //plugboard symetry
        for (int i = 0; i < alphabet.size(); ++i) {
            for (int j = 0; j < alphabet.size(); ++j) {
                BoolVar equal = Factory.isEqual(plug_board[i], j);
                cp.post(new IsEqual(equal, plug_board[j], i));
            }
        }

        // Assignment code text
        for(int i = 0; i < K; ++i){
            text[2 * M + 3][i].assign(alphabet.get_digit(code_char[i]));
        }
        //assignement crib
        for (int i = 0; i < crib_char.length; ++i) {
            text[0][i].assign(alphabet.get_digit(crib_char[i]));
        }
        //plug board constraints
        for (int i = 0; i < K; ++i) {
            cp.post(new Element1DVar(plug_board, text[0][i], text[1][i]));
            cp.post(new Element1DVar(plug_board, text[2 * M + 2][i], text[2 * M + 3][i]));
        }

        for (int i = 0; i < K; ++i) {
            for(int j = 0; j < M; ++j){
                int value = (int) (i / Math.pow(N, j));
                IntVar idx = Factory.modulo(Factory.plus(key[j], value), N);
                text[j+1+1][i] = Factory.element(transpo_matrix[j], idx, text[j+1][i]);
            }
            //reflector
            text[M + 1+1][i] = Factory.element(reflector, text[M+1][i]);
            for(int j = 0; j < M; ++j){
                int rotor_idx = M - (j + 1);
                int value = (int) (i / Math.pow(N, rotor_idx));
                IntVar idx = Factory.modulo(Factory.plus(key[rotor_idx], value), N);
                cp.post(new Element2D(transpo_matrix[rotor_idx], idx, text[M + 2 + j+1][i], text[M + 1 + j+1][i]));
            }
        }

        // implied constraints, only a few nodes are avoided with them, not a good tradeof
        /*
        for (int i = 0; i < code_char.length; ++i) {
            for (int j = i + 1; j < code_char.length; ++j) {
                if(i < crib_char.length && j < crib_char.length){
                    if (crib_char[i] == crib_char[j])
                        cp.post(new Equal(text[1][i], text[1][j] ));
                    else
                        cp.post(new NotEqual(text[1][i], text[1][j]));
                }
                if(i < crib_char.length) {
                    if (crib_char[i] == code_char[j])
                        cp.post(new Equal(text[1][i], text[2 * M + 2][j]));
                    else
                        cp.post(new NotEqual(text[1][i], text[2 * M + 2][j]));
                }
                if(j < crib_char.length) {
                    if (code_char[i] == crib_char[j])
                        cp.post(new Equal(text[2 * M + 2][i], text[1][j]));
                    else
                        cp.post(new NotEqual(text[2 * M + 2][i], text[1][j]));
                }

                if (code_char[i] == code_char[j])
                    cp.post(new Equal(text[2 * M + 2][j], text[2 * M + 2][i]));
                else
                    cp.post(new NotEqual(text[2 * M + 2][j], text[2 * M + 2][i]));
            }
        }
         */

        DFSearch dfs = makeDfs(cp, and(firstFail(key), firstFail(plug_board), firstFail(text[0])));

        IntVar[] plaintext = text[0];
        /*****************************************/

        dfs.onSolution(() -> {
                    System.out.print(" -");
                    for (int i = 0; i < N; ++i) {
                        System.out.print(plug_board[i]);
                    }
                    System.out.print(" / ");
                    for (int i = 0; i < M; ++i) {
                        System.out.print(key[i]);
                    }
                    System.out.print(" -> ");
                    alphabet.print(plaintext);
                    System.out.println();
                }
        );

        SearchStatistics stats = dfs.solve();
        System.out.println(stats);

    }
}

